(** backward policy iteration to compute gfp, simple version *)

open Syntax

(* structures *)

type rhs_lp =
	{ 
	   opt : Num.t ref;
           b_splx : Splx.t;
           lmat : (int * Vec.t * Num.t ref) list;
	   i_lmat : int;
	   splx : Splx.t ref;
 	   tp_l : Vec.t;
	   result : (Vec.t * Num.t) ref;
	   rhs_t : Vec.t;
           offset : int;
	   active : bool ref;
	}
	(* optimal sol, lp from transition, other lines, lp reconstructed,
	    template line, equation from dual solution,
				rhs_trans offset_templ 
	    active : non-empty splx *)
		(* mlglp_prob : max fx | ( A ) (x / x') <= (q / \rho) *)


(* change the variables of a Rtree, must go to utilities *)
(* f : Var.t -> Var.t is supposed to be one-to-one *)
let fold_with_var_rtree f isz res rt =
	let rec reverse_var v ret =
	   match v with
	     Var.XH -> ret
           | Var.XO tv -> reverse_var tv (Var.XO ret)
	   | Var.XI tv -> reverse_var tv (Var.XI ret)
	in
	let rec pc tres trt pref = 
	   match trt with
	     Rtree.Nil -> tres
	   | Rtree.Sub (l,n,r) -> let tres1 = pc tres l (Var.XO pref) in
		   	          let tres2 = if isz n then tres1
					else (f tres1 n 
					        (reverse_var pref Var.XH)) in
			          pc tres2 r (Var.XI pref)
	in pc res rt Var.XH
let transform_vec v f z isz inz =
	fold_with_var_rtree
		(fun rt vl va -> Rtree.set z rt (f va) vl) isz inz v
let primed_vec (v:Vec.t) nbvars =
	transform_vec v (fun va -> Var.fromInt (Var.toInt va+nbvars)) 
			Num.z (fun x -> Num.cmpz x = 0) Vec.nil

exception Simplex_RHS_Minf
let reconstruct_simplex splx lst_trans = 
	List.fold_left (fun spl (i,vc,fp) ->
				if not (Num.isReal !fp) then
					raise Simplex_RHS_Minf;
				let cst = { Cstr.typ = Cstr.Le;
					    Cstr.v = vc;
					    Cstr.c = !fp; } in
	                Configuration.debugf ~level:9 "[bwdPI,reconstruct_simplex]add %s@." (Cstr.pr (fun v -> "r"^(string_of_int (Var.toInt v))) cst);
			Splx.add spl (i,cst)) splx lst_trans

let use_witness witness max_temp_size rhs_trans nbtmp offset =
	  let p = ref (Vec.nil) 
	  and v = ref (Num.z) in
	  List.iter ( fun (i,vl) ->
	     if (i<=max_temp_size) then
	     begin
	       let nv = Var.fromInt (i+offset) in
	         assert (Num.cmpz (Vec.get (!p) nv) = 0);
	       p:= (Vec.set (!p) nv vl) 
	     end else
	     begin
		v:=Num.sub !v 
		     (Num.mult vl (Vec.get rhs_trans (Var.fromInt 
				          (i-max_temp_size))))
	     end) witness;
	  let vm = Var.fromInt nbtmp in
	  p := (Vec.set (!p) vm (Num.sub (Vec.get (!p) vm) Num.u));
	  (!p,!v)


(* loop : Cstr.t list list *)
(* postfp : Num.t array array (assuming we already know the values in the
   template) *)
(* template : Vec.t list option array (as vectors of non-primed variables ?) *)
(* array_splits : Cstr.t list array *)
(* resultat : rhs_trans array array * (rhs_trans list) array array *)
(** equation construction, from nb variables, program loop, postfp element, and
    template *)
let construct_equation_set nbvar loop postfp template array_splits = 
	(* construct a list, with one array for each transition
	   separately (some mlglp_prob are identical) *)
	(* the first element of the list gives a global minimum *)
	(* TODO : hyp : no element of postfp is -oo or +oo *)
	let max_vartemp = ref 0 in
	let nb_splits = Array.length template in
	Configuration.debugf "[bwdPI] construct_equation_set %d %d@." 
				nbvar nb_splits;
	let eqs_results = Array.init (Array.length postfp) (fun i -> 
			   let lvtemp = Array.length postfp.(i) in
			   if (lvtemp > !max_vartemp) then
				max_vartemp:= lvtemp;
			   Array.make lvtemp []) 
			in
	List.iteri 
		(fun num_trans trans -> 
			(* each transition is a list  of contraints *)
			(* construction of (   Q    )
					   ( S1   0 ) 
					   ( 0   T2 ) as
				a list of constraints *)
		   let nbtemp = ref (!max_vartemp) in 
					(* FIXME : template_size ? *)
		   let rhs_trans = ref Vec.nil in
			(* construction of (   Q    ) *)
		   begin try
		   let splx = List.fold_left (fun spl t ->
                                       nbtemp := !nbtemp+1;
				       rhs_trans := Vec.set (!rhs_trans)
                                             (Var.fromInt 
						(!nbtemp- !max_vartemp))
                                             (Cstr.get_c t);
				       Splx.add spl ((!nbtemp), t))
				(Splx.mk (Var.fromInt (2*nbvar+1)) []) trans in
							(* FIXME : nbvar *)
		   let offset_split = ref 1 in
		   Array.iteri (fun num_split split ->
		     match template.(num_split) with
		       None -> ()
		     | Some temp_num_split ->
		       begin
			  let nbtemp2 = ref !nbtemp in
			  let rhs_trans2 = ref !rhs_trans in
			  (* adding ( S1  0 ) after  *)
			  begin try
			  let splx2 = List.fold_left (fun spl t ->
                                         nbtemp2 := !nbtemp2+1;
				         rhs_trans2 := Vec.set (!rhs_trans2)
                                            (Var.fromInt 
						(!nbtemp2- !max_vartemp))
                                             (Cstr.get_c t);
				       Splx.add spl ((!nbtemp2), t))
				splx split in
			  let offset_templ = ref 0 in
			  Array.iteri (fun num_templ templ_opt ->
			    match templ_opt with
			      Some templ ->
				begin
				(* now we add ( 0 T_i ) *)
			        let lmat = List.mapi
			            (fun i vc ->
					(i+1),primed_vec vc nbvar,
					ref (postfp.(num_templ).(i)))
							templ in
				
				begin try
				let nsplx = 
					ref (reconstruct_simplex splx2 lmat)
			        in 
	                        Configuration.debugf ~level:9 
				    "[bwdPI] constructed complete splx (tr = %d, split = %d templ = %d) [%s]@." num_trans num_split num_templ
				(Splx.pr (fun v -> "r"^
					(string_of_int (Var.toInt v))) !nsplx);
			        ignore (Splx.check !nsplx);
			        Configuration.debugf ~level:9 
					"[bwdPI] splx checked, feasible@.";
				let active = ref true in
			        List.iteri ( fun nbt vect_template -> 
				   Configuration.debugf ~level:9 "[bwdPI] check optimal, objective %s : " (Vec.pr (fun v -> "r"^(string_of_int (Var.toInt v))) vect_template);
				   let rlp =
					  {
					     opt = ref (Num.of_float infinity);
					     b_splx = splx2;
					     lmat = lmat;
					     i_lmat = num_templ;
					     splx = nsplx;
					     tp_l = vect_template;
					     result = ref (Vec.nil,Num.z);
					     rhs_t = !rhs_trans2;
				             offset = !offset_templ;
					     active = active;
					  }
				   in
				   begin
				   match Splx.Opt.max !nsplx vect_template with
				     Splx.Opt.Finite (vl,witness) 
				   | Splx.Opt.Sup (vl,witness) ->
					(* finite or sup does not matter *)
					begin
					  Configuration.debugf ~level:9 "result %s@." (Num.pr vl);
					  let (p,v) =
				             use_witness witness
						!max_vartemp !rhs_trans2
						(!offset_split+nbt) 
						!offset_templ in
					  rlp.opt := vl;
					  rlp.result := (p,v);
					end
				   | Splx.Opt.Infty ->
					begin
			                  Configuration.debugf ~level:9
					    "result +oo@.";
					end
				  end;
				  eqs_results.(num_split).(nbt) 
					     <- rlp
					      ::eqs_results.(num_split).(nbt)
			        ) temp_num_split
				with
				     Simplex_RHS_Minf
				   | Splx.Unsat (_) -> 
					Configuration.debugf ~level:9 "[bwdPI] unfeasible (tr = %d, split = %d templ = %d)@." num_trans num_split num_templ
				end;
				offset_templ := !offset_templ + 
					List.length templ
			      end
			    | None -> ()
			  ) template
			  with
			     Splx.Unsat (_) -> Configuration.debugf ~level:9 "[bwdPI] unfeasible (tr = %d, split = %d)@." num_trans num_split
			  end;
			  offset_split := !offset_split + 
					Array.length postfp.(num_split)
			end
		    ) array_splits 
		    with
			 Splx.Unsat (_) -> Configuration.debugf ~level:9 "[bwdPI] unfeasible (tr = %d)@." num_trans
		    end
	) loop;
	!max_vartemp, eqs_results

(* eqs : return of construct_equation_set : (rhs_lp list) array array *)
(* pfp : new pfp :  Num.t array array *)
(** improving the strategy, recomputing the linear programs *)
let improve_strat eqs pfp max_vtemp =
	Configuration.debugf "[bwdPI] improve_strat@.";
	let nbvar = ref 0 in
	Array.iteri (fun isplit syst ->
	   for itempl = 0 to Array.length syst-1 do
	      nbvar := !nbvar +1;
	      if (Num.isReal pfp.(isplit).(itempl)) then
	      begin
	      let eql = syst.(itempl) in
	      Configuration.debugf "[bwdPI,improve_strat] var %d (spl=%d tpl=%d), sol = %s@." !nbvar isplit itempl (Num.pr pfp.(isplit).(itempl));
	      List.iteri (fun itrans rlp ->
		   if (itempl=0) && !(rlp.active) then
		   begin
		     begin try
		     (* we reconstruct the linear program *)
		       List.iteri (fun i (_,_,e) -> e := pfp.(rlp.i_lmat).(i))
					 rlp.lmat;
		       rlp.splx := (reconstruct_simplex rlp.b_splx rlp.lmat);
		       ignore (Splx.check !(rlp.splx));
	               Configuration.debugf ~level:8 "[bwdPI,improve_strat]simplex %d (%d) feasible@." isplit itrans
		     with
		       Simplex_RHS_Minf
		     | Splx.Unsat (_) -> Configuration.debugf ~level:8 "[bwdPI,improve_strat]simplex %d (%d) unfeasible@." isplit itrans ; 
				         rlp.active := false
		     end	
		   end;
		   if !(rlp.active) then
		   begin
		     match Splx.Opt.max !(rlp.splx) rlp.tp_l with
		       Splx.Opt.Finite (vl,witness) 
		     | Splx.Opt.Sup (vl,witness) -> 
			  Configuration.debugf ~level:9 "[bwdPI,improve_strat] value computed var=%d trans=%d : %s@." !nbvar itrans (Num.pr vl);
			 if (Num.cmp vl !(rlp.opt) < 0) then
			 begin
			   let (p,v) =
		             use_witness witness
				max_vtemp rlp.rhs_t
				!nbvar rlp.offset in
			   rlp.result := (p,v);
			   rlp.opt := vl;
			 end
		     | Splx.Opt.Infty ->
			 begin
			   (* FIXME check !fr = infty *)
			   rlp.opt := (Num.of_float infinity);
			 end
		   end
	      ) eql
	      end else syst.(itempl) <- []
	   done
	) eqs;
	let global_changed = ref false in
	Array.iteri (fun isplit syst ->
		for j=0 to Array.length syst -1 do
		    if syst.(j) <> [] then
		    begin
		      syst.(j) <- List.filter (fun a -> !(a.active)) syst.(j);
		      if (not !global_changed) && 
		         (List.for_all (fun a -> 
				Num.cmp !(a.opt) pfp.(isplit).(j) < 0) 
				syst.(j)) then
			  global_changed:=true
		    end
		done) eqs;
	!global_changed

(* return : int : nombre de variables (version finie)
            Vec.t array : liste de chaque contrainte (template_size)
	    Var.t : horizon
	    bool array : contraintes actives 
            Vec.t contenant l'objectif (version finie) *)

type local_strat =
	{ nb_lambdas : int;
	  horizon : Var.t;
	  matrix : Vec.t array;
	  active : bool array;
	  objective : Vec.t;
	}
          

(* max Cst.lambda | lambda<=0 et VECS . lambda = (1) *)
(* eqs : rhs_trans list array array ; tfp : Num.t array array *)
(** construct local strategy equation system using the global system *)
let construct_local_strat eqs template_size tfp =
	(* eqs : rhs array list *)
	Configuration.debugf "[bwdPI] construct_local_strat, nbv=%d@."
				template_size;
	let nb_lambdas = ref 0 in
	let var_lambdas = ref Var.XH in
	let vecs = Array.make template_size Vec.nil in
	let vecs_active = Array.make template_size true in
	let objectif = ref Vec.nil in
	let num_var = ref 0 in
	for num_split = 0 to Array.length tfp-1 do
           for num_temp = 0 to Array.length tfp.(num_split)-1 do
	      Configuration.debugf ~level:5 
		"[bwdPI construct_local_strat] variable %d (split %d) (temp %d)@." !num_var num_split num_temp;
	      let u = tfp.(num_split).(num_temp) in
	      if List.exists (fun a -> Num.cmp !(a.opt) u >=0) 
					eqs.(num_split).(num_temp) then
	      begin
		Configuration.debugf ~level:5 "[bwdPI construct_local_strat] keep pfp (val = %s)@." (Num.pr u);
		nb_lambdas := !nb_lambdas+1;
		let vr = !var_lambdas in
		var_lambdas := Var.next (!var_lambdas);
		vecs.(!num_var) <- Vec.set (vecs.(!num_var)) vr (Num.negU);
		objectif := Vec.set (!objectif) vr (Num.neg u)
	      end
	      else
	      begin
		(* if eqs.(num_split).(num.temp) is empty, the value is -oo *)
		if (eqs.(num_split).(num_temp) = []) then
		begin
		   Configuration.debugf ~level:5 "[bwdPI construct_local_strat] -oo value@.";
		   vecs_active.(!num_var) <- false
		end
		else
		List.iter (fun a ->
		   nb_lambdas := !nb_lambdas+1;
		   let vr = !var_lambdas in
		   var_lambdas := Var.next (!var_lambdas);
		   Configuration.debugf ~level:5 "[bwdPI construct_local_strat] local val %s@." (Num.pr !(a.opt));
	           let (v,u) = !(a.result) in
		   for j=0 to template_size-1 do
		       let uv = Vec.get v (Var.fromInt (j+1)) in
		       if (Num.cmpz uv <> 0) then
		          vecs.(j) <- Vec.set (vecs.(j)) vr uv
		   done;
		   objectif := Vec.set (!objectif) vr u
		) eqs.(num_split).(num_temp);
	      end;
	      num_var := !num_var +1
	    done
	done;
	Configuration.debugf "[bwdPI construct_local_strat] total equations: %d@." !nb_lambdas;
	{ nb_lambdas = !nb_lambdas;
	  horizon = !var_lambdas;
	  matrix = vecs;
	  active = vecs_active;
	  objective = !objectif;
	}

let array_fold_left_option f a b =
	Array.fold_left (fun x y -> match y with
				      None -> x
				    | Some e -> f x e) a b

exception SolEmpty
let solve_local_strat strat p_fp =
       (* first thing : produce the infinity lp *)
	Configuration.debugf "[bwdPI] solve_local_strat, nblambdas=%d@." 
					strat.nb_lambdas;
	let template_size = Array.length strat.matrix in
	let nb_cstrs_oo = ref 0 in
	let obj = ref Vec.nil in
	let lambdas_oo = ref strat.horizon in
(*	let real_values = Array.make template_size true in
	let nb_reals = ref template_size in (* FIXME : deal with +oo *) *)
	let cstrs_oo = 
		Array.mapi (fun i vec ->
			if strat.active.(i) then
			begin
			  let vr = !lambdas_oo in
			  lambdas_oo := Var.next (!lambdas_oo);
			  nb_cstrs_oo := !nb_cstrs_oo +1;
			  let v1 = Vec.set vec vr Num.negU in
			  obj := Vec.set (!obj) vr Num.u;
			  Some ((i+1),
			        { Cstr.typ = Cstr.Eq; 
                                  Cstr.v = v1; 
                                  Cstr.c = Num.u })
			end else None
			) strat.matrix in
	if (!nb_cstrs_oo = 0) then raise SolEmpty;
	let nb_lambdas_oo = strat.nb_lambdas + !nb_cstrs_oo in
	let plambdas = ref Var.XH in
	let cstrs_oo2 = Array.init nb_lambdas_oo 
			(fun i -> let vr = !plambdas in
				  plambdas := Var.next !plambdas;
	                          (i+1)+template_size,
				  { Cstr.typ = Cstr.Le;
				    Cstr.v = Vec.mk [Num.u, vr];
				    Cstr.c = Num.z }) in
	assert (!plambdas = !lambdas_oo);
	Configuration.debugf ~level:8 "[bwdPI,solve_local_strat] Constructing splx_oo (nb_lambdas_oo = %d)...@." nb_lambdas_oo;
	begin
	try
	    let splx_oo_1 = Splx.mk (!lambdas_oo) [] in
	    let splx_oo_2 = Array.fold_left (fun spl icstr -> 
					Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] add %s [%d]]@." (Cstr.pr (fun u -> "l"^(string_of_int (Var.toInt u))) (snd icstr)) (fst icstr);
					Splx.add spl icstr)
					splx_oo_1 cstrs_oo2 in
	    Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...variables bounds added@.";
	    let splx_oo = array_fold_left_option (fun spl icstr -> 
				Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] add %s [%d]@." (Cstr.pr (fun u -> "l"^(string_of_int (Var.toInt u))) (snd icstr)) (fst icstr);
				Splx.add spl icstr)
			    	splx_oo_2 cstrs_oo in
	  Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...basic constraints added@.";
	    match Splx.Opt.max splx_oo (!obj) with
	      Splx.Opt.Infty -> failwith "splx_oo is unbounded ???"
	    | Splx.Opt.Sup (_) -> failwith "splx_oo returns Sup ???"
	    | Splx.Opt.Finite (tot,witness) -> 
	                Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...max value : %s@." (Num.pr tot);
			if (Num.cmpz tot > 0) then
			List.iter (fun (n,v) ->
	                        Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...witness : %d,%s@." n (Num.pr v);
				if (n<=template_size) && (Num.cmpz v >0) then
				begin
				   strat.active.(n-1) <- false
				end
				) witness
	with
	     Splx.Unsat (w) -> begin
				 Configuration.errorf "splx_oo is unfeasible (witness = [%a]) ???@." (fun fmt a -> List.iter (fun (n,v) ->
			Format.fprintf fmt "%s.e%d " (Num.pr v) n) a) w;
				 failwith "splx_oo is unfeasible ???"
			       end
	end;
	let nb_cstrs = ref 0 in
(*	let real_values = Array.make template_size true in
	let nb_reals = ref template_size in (* FIXME : deal with +oo *) *)
	let cstrs = 
		Array.mapi (fun i vec ->
			if strat.active.(i) then
			begin
			  nb_cstrs := !nb_cstrs +1;
			  Some (i+1,
			  { Cstr.typ = Cstr.Eq; 
                            Cstr.v = vec; 
                            Cstr.c = Num.u })
			end else None
			) strat.matrix in
	if (!nb_cstrs = 0) then raise SolEmpty;
	let plambdas = ref Var.XH in
	let cstrs_2 = Array.init strat.nb_lambdas 
			(fun i -> let vr = !plambdas in
				  plambdas := Var.next (!plambdas);
	                          i+1+template_size,
				  { Cstr.typ = Cstr.Le;
				    Cstr.v = Vec.mk [Num.u, vr];
				    Cstr.c = Num.z }) in
	assert (!plambdas = strat.horizon);
	Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] Constructing splx.@.";
        let result = Array.make template_size (Num.z) in
	begin
	try
	   let splx_1 = Splx.mk !plambdas [] in
	   let splx_2 = Array.fold_left (fun spl icstr -> Splx.add spl icstr)
					splx_1 cstrs_2 in
	   Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...variables bounds added@.";
	   let splx = array_fold_left_option
					(fun spl icstr -> Splx.add spl icstr)
					splx_2 cstrs in
	   Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...basic constraints added@.";
	   match Splx.Opt.max splx strat.objective with
	      Splx.Opt.Infty -> failwith "splx is unbounded ???"
	    | Splx.Opt.Sup (_) -> failwith "splx returns Sup ???"
	    | Splx.Opt.Finite (_,witness) -> 
			List.iter (fun (n,v) ->
	                        Configuration.debugf ~level:9 "[bwdPI,solve_local_strat] ...witness : %d,%s@." n (Num.pr v);
				if (n<=template_size) && (Num.cmpz v <>0) then
				   result.(n-1) <- v
				) witness
	with
	     Splx.Unsat (_) -> failwith "splx is unfeasible ???"
	end;
	let pc_var = ref 0 in
	Array.map (fun a ->
		Array.map (fun _ -> 
			let ppc = !pc_var in
			pc_var := ppc +1; 
			if strat.active.(ppc) then
				result.(ppc) else (Num.of_float (-.infinity)))
			a) p_fp

let print_template fmt tmp =
	 Array.iteri (fun it template ->
		    if it > 0 then Format.pp_print_string fmt ";";
		    begin
		    match template with
		      None -> Format.pp_print_string fmt "bot"
		    | Some t ->
		       Format.pp_print_string fmt "[";
		       List.iteri (fun ix x -> 
				if ix>0 then Format.pp_print_string fmt "|";
				Format.pp_print_string fmt 
                                 (Vec.pr (fun i -> "l"^(string_of_int
					   (Var.toInt i))) x)) t;
		       Format.pp_print_string fmt "]"
		    end) tmp 

let print_sol fmt tfp = 
	Array.iteri (fun it a ->
		if it > 0 then Format.pp_print_string fmt ";";
		Format.pp_print_string fmt "[";
	        Array.iteri (fun i b ->
			if i>0 then Format.pp_print_string fmt ";";
			Format.pp_print_string fmt (Num.pr b)) a;
		Format.pp_print_string fmt "]") tfp

(* general policy iteration *)
(* loop : Cstr.t list list*)
(* template_base : Vec.t list option array *)
(* postfixpoint : Num.t list array *)
(* array_splits : Cstr.t list array *)
let gfp_PI nbv loop template_base postfixpoint array_splits =
	let nb_it = ref 1 in
	let template_size = ref 0 in
	let template_fp =
		Array.map (fun t ->
			let a = Array.of_list t in
			template_size := !template_size + Array.length a;
			a) postfixpoint
	in
        Configuration.debugf ~level:7 
			"[bwdgfp PI] postfp = [%a] ; template = [%a]@."
			print_sol template_fp
			print_template template_base;
	let sol = ref template_fp in
	begin
	try
	  Array.iter (Array.iter (fun x -> if not (Num.isReal x) then 
				failwith "gfp_PI non finite initial pfp"))
			template_fp;
	  let max_templ,eqs = 
		construct_equation_set nbv loop template_fp template_base
						array_splits in
	  let unstable = ref true in
	  while (!unstable) do 
	     Configuration.debugf "[bwdgfp PI] : iteration %d (sol %a)@." (!nb_it) print_sol (!sol);
	     let strat = construct_local_strat eqs !template_size template_fp in
	     Configuration.debugf "[bwdgfp PI] : strat constructed (nblambdas = %d, vecs = %a, obj = %s)@." strat.nb_lambdas 
                (fun fmt a -> Array.iteri (fun i vc -> 
				Format.fprintf fmt "(%d) %s ; " (i+1) 
				  (Vec.pr (fun l -> "l"^
					(string_of_int (Var.toInt l))) vc)) a)
		strat.matrix
		(Vec.pr (fun l -> "l"^ (string_of_int (Var.toInt l))) 
			strat.objective);
(*	     Configuration.debugf "[bwdgfp PI] : end solving strat (sol %a)@." (fun fmt a -> Array.iter (fun elt -> Format.fprintf fmt "%g " elt) a) sol; *)
	     sol:= solve_local_strat strat template_fp;
	     unstable := improve_strat eqs (!sol) max_templ; 
	     nb_it := !nb_it + 1
          done;
	  Some (!sol)
	with
	  SolEmpty -> begin
			Configuration.debugf "[bwdgfp PI] : empty return@.";
			None
		      end
	end
