(* $Id: bwdsem.ml,v 1.4.2.8 2013/10/02 15:29:40 dmasse Exp $ *)

(** computation of the backward semantics *)

open Syntax

(** compute an overapproximation of the greatest fixpoint *)
let backward_sem bwddom reachable pg =
	Configuration.debugf "[bwdsem] Backward analysis@.";
	let _,split,loop = pg in
	let current_state = 
		begin
		    match reachable with
			  None -> ref (bwddom#top)
			| Some pol1 -> begin
				        (bwddom#affect_linconsts_to_split pol1);
					ref (pol1)
				       end
		end in
	let loop_conv = List.map bwddom#convert_trans loop in
	let stable = ref false in
	let i = ref 0 in
	while ((not !stable) && (!i<Configuration.direct_gfp_steps#get)) do
		Configuration.debugf "[bwdsem] Step %d in descending sequence@."
				(!i);
		let new_states = List.map 
			(fun trans -> bwddom#apply_trans_bwd
				 !current_state trans) 
				loop_conv
		in
		let pre_state = bwddom#join_list new_states in
		Configuration.debugf "[bwdsem] pre state (after glb) : %a@."
				bwddom#print_state pre_state;
		stable := bwddom#leq_state !current_state pre_state;
		if (not (!stable)) then
		begin
		  if ((!i+1) mod Configuration.bwdPI_frequency#get = 0) then
		  begin
		    Configuration.debugf "[bwdsem] using PI@.";
		    Configuration.debugf 
			"[bwdsem] pre state (before PI) : %a@."
				bwddom#print_state pre_state;
		    let tbase,pfp = bwddom#state_to_template pre_state in
		    let npfp = BwdPI.gfp_PI (!nb_variables) loop_conv
					 tbase pfp bwddom#get_array_splits in
		    let resPI = bwddom#template_to_state tbase npfp in
		    Configuration.debugf 
			"[bwdsem] pre state (after PI) : %a@."
				bwddom#print_state resPI;
		    current_state := resPI;
		  end else   
		       current_state := pre_state;
		  i := !i+1
		end
	done;
(*	 let result = bwddom#translate_state !current_state in *)
	(* printing *)
	Configuration.outf 
		"[bwdsem] Backward analysis finished. Potentially non-terminating states:\n%a\n"
		bwddom#print_state (!current_state);
	!current_state

