(** configuration.ml : options file *)

(** Forward analysis *)

class ['a] config_option (def: 'a) name (argset : ('a -> unit) -> Arg.spec)
				docstring =
object (self)
	val mutable x = def
	val option_name = "-"^name
	val argset = argset
	val docs : string = docstring def

	method get = x
	method set y = x <- y

	method getarg =
		option_name,(argset self#set),docs
end

let options_arg = ref []

let add_options_arg a =
	options_arg := a::(!options_arg)

let int_config_option def name docstring =
	let a = new config_option def name (fun f -> Arg.Int f) 
				(Printf.sprintf docstring)
	in
	add_options_arg (a#getarg);
	a
let bool_config_option def name docstring =
	let a = new config_option def name (fun f -> Arg.Bool f) 
				(Printf.sprintf docstring)
	in
	add_options_arg (a#getarg);
	a
let fmt_config_option def name docstring =
	let a = new config_option def name 
		(fun f -> Arg.String 
		   (fun s -> f (Format.formatter_of_out_channel (open_out s))))
	        (fun _ -> Printf.sprintf docstring)
	in
        add_options_arg (a#getarg);
        a


(** widening/narrowing limits *)
let widening_limit = int_config_option 3 "fwd_widen" 
	 "fwd: increasing sequence before widening or PI (default %d)"
let narrowing_steps = int_config_option 1 "fwd_narrow"
	 "fwd: narrowing sequence (default %d)"

(** Forward analysis : use PI instead of widening *)
let fwd_policy_iteration = bool_config_option true "fwd_PI"
	 "fwd: use PI instead of widening (default %B)"
(** maximum number of iterations on fwd PI *)
let fwd_PI_max_iterations = int_config_option 5 "fwd_PI_maxit"
	 "fwd: max number of iterations for each PI (default %d)"
(** lookahead ``widening'' for PI: ignore transitions which are ineffective
    when we start using PI *)
let fwd_PI_lookahead = bool_config_option true "fwd_PI_lookahead"
	 "fwd: \"lookahead\" PI (default %B)"

(** Backward analysis *)

(** gfp approximation *)
(** total number of iterations *)
let direct_gfp_steps = int_config_option 10 "bwd_steps"
	 "bwd: max number of decreasing steps for gfp (default %d)"
(** When to use PI *)
let bwdPI_frequency = int_config_option 4 "bwd_PI_freq"
 	 "bwd: frequency of PI for gfp (default %d)"

(** Reading/writing *)

(* outf *)
let out_fmt = fmt_config_option Format.std_formatter "out"
	 "output file (defaut stdout)"
let outf ?(ignore=false) f = 
	(if ignore then Format.ifprintf else Format.fprintf) (out_fmt#get) f

(* debugf *)
let debug_level = int_config_option 0 "debug"
	 "debug level (default %d)"
let debugf ?(level=1) f = outf ~ignore:(level > debug_level#get) f
let debugdo ?(level=1) action = if level <= debug_level#get then action ()

(** errorf *)
let err_fmt = fmt_config_option Format.err_formatter "err"
	 "error file (default stderr)"
let errorf f = Format.fprintf (err_fmt#get) f

(*
(** epsilon : should be replaced by an estimation of the safety of the 
    computation *) 
let epsilon = 1e-6 (* not a good value : approximation should be much better *)

*)
