(* $Id: fwdsem.ml,v 1.5.2.6 2013/09/27 09:20:58 dmasse Exp $ *)

(** computation of the forward semantics *)

open Syntax

(** compute an overapproximation of reachability analysis *)
let forward_sem fwddom et pg =
	Configuration.debugf "[fwdsem] Forward analysis@.";
	let init,split,loop = pg in
	let current_state = ref (match et with
			           None -> fwddom#top
				 | Some a -> a) in
	let init_conv = fwddom#convert_init init in
	let loop_conv = List.map fwddom#convert_trans loop in
        current_state := fwddom#apply_init !current_state init_conv;
	Configuration.debugf ~level:5 "[fwdsem] initial state : %a@."
				fwddom#print_state !current_state;
	let init_state = !current_state in
	let stable = ref false in
	let postfp = ref false in
	let i = ref 0 in
	while (not (!postfp) || 
	    (not (!stable) && (!i<Configuration.narrowing_steps#get))) do
		Configuration.debugf "[fwdsem] Step %d in %s sequence@."
				(!i) (if (!postfp) then "descending"
						   else "ascending");
		let new_states = List.map 
			(fun trans -> fwddom#apply_trans !current_state trans) 
				loop_conv
		in
		let final_state = fwddom#join_list (init_state::new_states) in
		Configuration.debugf ~level:5 "[fwdsem] after transition : %a@."
				fwddom#print_state final_state;
		stable := (if !postfp then
				fwddom#leq_state !current_state final_state
				else 
				fwddom#leq_state final_state !current_state);
		Configuration.debugf ~level:9 "[fwdsem] stable : %b@." !stable;
		if (not (!stable)) then
		begin
		  i := !i+1;
		  if ((not !postfp) && 
			(!i > Configuration.widening_limit#get)) then
		  begin
(* 		  No FwdPI for now *) 
(*		       if (Configuration.fwd_policy_iteration#get) then
		       begin
			 Configuration.debugf "[fwdsem] using PI@.";
                         fwddom#approximate_state final_state;
                         Configuration.debugf 
                           "[fwdsem] approximated post state (before PI) : %a@."
                                fwddom#print_state final_state;
			 let tfinal_state = 
				fwddom#translate_state final_state in
			 let temp,pfp = 
				construct_template_from_state tfinal_state in
			 let npfp = 
				FwdPI.lfp_fwdPI (!nb_variables) 
						loop pfp temp in
			 let resPI = 
				return_state_from_template temp npfp in
			 Configuration.debugf 
                          "[fwdsem] post state (after PI, not translated) : %a@."
                                print_etat resPI;
			 current_state := fwddom#convert_state resPI;
                         Configuration.debugf 
				"[fwdsem] post state (after PI) : %a@."
                                fwddom#print_state (!current_state);
			 (* if PI finished, we got a fixpoint, however,
			    PI may have a max number of iterations or
			    use lookahead. Then we
			    reinitialise the increasing sequence *)
			 (* FIXME : the sequence may not terminate?? *)
			 i := 0
		       end else *)
		       begin
		         let crt = fwddom#join_state 
				!current_state final_state in
		         current_state := fwddom#widen_state 
					  !current_state crt;
		         Configuration.debugf ~level:5 
					"[fwdsem] widening : %a@."
				fwddom#print_state (!current_state);
		       end
		  end else current_state := final_state
		end else if (not !postfp) then
		begin
		     stable := fwddom#leq_state !current_state final_state;
		     postfp := true;
		     i := 0;
		end
	done;
(*	let result = fwddom#translate_state !current_state in *)
	(* printing *)
	Configuration.outf "[fwdsem] Forward analysis finished. Reachable states:\n%a\n"
		fwddom#print_state (!current_state);
	!current_state
