(** lexer.mll : lexer *)

(* $Id: *)

{
open Parser
exception Eof

let keyword_table = [ "Var", Var_Token ; "Init", Init_Token;
			"Loop", Loop_Token ; "Split", Split_Token ]
}

rule token = parse
  ("--!TRACE"[^'\n']*)          { token lexbuf }
| ("//"[^'\n']*)                { token lexbuf }
| ((['a'-'z'] | ['A'-'Z'])(['_' '|' '.']?(['a'-'z'] | ['A'-'Z'] | ['0'-'9']))*) 
as id				{ try List.assoc id keyword_table with
				  Not_found -> Identifier_Token id }
| ("'")				{ Apostrophe_Token }
| ('=')				{ Equal_Token }
| (">=")			{ GreatEq_Token }
| ("<=")			{ LessEq_Token }
| ('>')				{ Great_Token }
| ('<')				{ Less_Token }
| ('(')				{ LeftPar_Token }
| (')')				{ RightPar_Token }
| ('[')				{ LeftBra_Token }
| (']')				{ RightBra_Token }
| ('+')				{ Plus_Token }
| ('-')				{ Minus_Token }
| ('*')				{ Times_Token }
| ('/')				{ Div_Token }
| (';')				{ Semicolon_Token }
| (',')				{ Comma_Token }
| ['0'-'9']+'.'['0'-'9']+       { (Double_Token (Lexing.lexeme lexbuf)) }
| ['0'-'9']+       { (Integer_Token (Lexing.lexeme lexbuf)) }
| [' ' '\t' '\n' '\r' ] { token lexbuf }
| eof                   { Eof_Token }

