(* $Id: program.ml,v 1.4.2.3 2013/09/26 18:51:25 dmasse Exp $ *)

(** program.ml : main program for now *)

open Syntax
open Parser

exception Syntax_Error
let print_error lexbuf = 
     let curr_p = lexbuf.Lexing.lex_curr_p in
		Configuration.errorf "curr_p : %s %d %d %d@."
				curr_p.Lexing.pos_fname 
				curr_p.Lexing.pos_lnum
				curr_p.Lexing.pos_bol
				 curr_p.Lexing.pos_cnum

let parse_program c =
     let lexbuf = Lexing.from_channel c in
     try
            Parser.n_Entry Lexer.token lexbuf
     with
     | Failure s -> Configuration.errorf "Failure to read program: %s@." s;
                    raise Syntax_Error
     | Lexer.Eof -> Configuration.errorf "lexical error : ";
                    print_error lexbuf;
                    raise Syntax_Error
     | Parsing.Parse_error -> Configuration.errorf "syntax error : ";
			      print_error lexbuf;
                              raise Syntax_Error

let parse_args () =
	let file_to_parse = ref stdin in
	Arg.parse (!Configuration.options_arg)
		  (fun s -> file_to_parse := open_in s)
			"pater <options> <file>\n  <file> : file to analyse (default: stdin)"
	; !file_to_parse

let _ = (* Mlglpk.term_hook := 
		Some (fun s -> Configuration.debugf ~level:6 
				"[GLPK] %s@?" s; true);
	Mlglpk.glp_register_term_hook true; *)
	let prog = parse_program (parse_args ()) in 
			Configuration.debugf ~level:5 "Program finished.@.";
			Configuration.debugf "%a@." print_prog prog;
			let (_,splt,_) = prog in
			let domain = new Domain.domain (!Syntax.nb_variables)
							 splt
							 Syntax.get_var_name_p
			in 
			let r_fwd = Fwdsem.forward_sem domain None prog in
			ignore (Bwdsem.backward_sem domain (Some r_fwd) prog)
